import { EventEmitter } from './helpers/EventEmitter'
import { SatiError } from './SatiError'
import { SatiSocket, SocketState } from './SatiSocket'
import { events, methods, tasks, Task } from './types'

/**
 * high level api wrapper
 * @example
 * const sati = new Sati({
 * 	token: 'your token here'
 * })
 * await sati.init() // you must call the init method after construction
 * 
 * const task = await sati.solve('Turnstile', {
 * 	siteKey: '0x4AAAAAAAHMEd1rGJs9qy-0',
 * 	pageUrl: 'https://polygon.sati.ac/Turnstile'
 * })
 * 
 * console.log(task, task.result.token)
 * 
 * sati.close() // you must call close method after you've done 
 */
export class Sati extends EventEmitter<events> {
	private socket: SatiSocket
	private awaitedTasks: {
		[ index: number ]: {
			resolve(data: any): void,
			reject(reason: any): void
		}
	} = Object.create(null)
	
	/** @param token your api token. get it at https://sati.ac/dashboard */
	constructor({ token }: { token: string }) {
		super()
		this.socket = new SatiSocket(token)
		this.socket.on('event', ({ type, data }) => {
			this.emit(type as keyof events, data)
		})
		this.socket.on('stateChange', ({ state, error }) => {
			switch(state) {
				case SocketState.connected:
					// as of socket was disconnected, there might be missed events
					// so we need to recheck the state of each awaited task
					for(const [ id, awaited ] of Object.entries(this.awaitedTasks)) {
						this.call('getTask', { id: +id }).then(this.handleTaskResult.bind(this), err => {
							awaited.reject(err)
							delete this.awaitedTasks[+id]
						})
					}
				case SocketState.unrecoverable:
					// socket got an unrecoverable error, so it will never be connected again
					for(const awaited of Object.values(this.awaitedTasks)) {
						awaited.reject(error)
					}
			}
		})
		this.on('taskUpdate', this.handleTaskResult.bind(this))
	}

	private handleTaskResult(task: Task) {
		if(!(task.id in this.awaitedTasks)) return
		if(task.state !== 'error' && task.state !== 'success') return
		const awaited = this.awaitedTasks[task.id]

		if(task.state === 'error') {
			awaited.reject(new SatiError(`unable to solve ${task.type} task #${task.id}`))
		} else if(task.state === 'success') {
			awaited.resolve(task)
		}
		delete this.awaitedTasks[task.id]
	}

	/** connects to the socket, must be called before any other methods */
	public async init() {
		await this.socket.connect()
	}

	/** closes the socket */
	public close() {
		this.socket.close()
	}
	
	/** calls the api method, use top-level wrappers for them instead */
	public async call<M extends keyof methods>(method: M, data: methods[M]['request']): Promise<methods[M]['response']> {
		const resp = await this.socket.send('call', { method, data })

		if(!resp.success) {
			throw new SatiError(resp.data.description)
		}

		return resp.data
	}

	/**
	 * creates task, then waits for the result
	 * @throws {SatiError} if unable to solve
	 */
	public async solve<T extends keyof tasks>(type: T, data: tasks[T]['params']): Promise<Task<T, 'success'>> {
		const task = await this.call('createTask', { type, data })

		return new Promise((resolve, reject) => {
			this.awaitedTasks[task.id] = { resolve, reject }
		})
	}

	public async getBalance() {
		return (await this.call('getBalance', {})).balance
	}
}
