import { EventEmitter } from './helpers/EventEmitter'
import { SocketAdapter } from '%env%'
import { SatiError } from './SatiError'
import { endpoint } from './config'

interface WebSocketMessage {
	id: number
	type: string
	data: any
	to?: number
}

export enum SocketState {
	connected,
	disconnected,
	unrecoverable
}

/** low-level api wrapper */
export class SatiSocket extends EventEmitter<{
	event: {
		type: string,
		data: any
	},
	stateChange: {
		state: SocketState,
		error: Error | null
	}
}> {
	private token: string
	private socket: SocketAdapter | null = null
	private error: Error | null = null
	private state: SocketState = SocketState.disconnected

	private idCounter = 0
	private awaitedMessages: Record<number, {
		resolve: (data: any) => void
		reject: (error: any) => void		
	}> = Object.create(null)
	private queue: [(data: any) => void, string, any][] = []

	constructor(token: string) {
		super()
		this.token = token
	}

	private setState(state: SocketState, error: Error | null = null) {
		this.state = state
		this.error = error
		this.emit('stateChange', { state, error })
	}

	public async connect() {
		if(this.socket || this.state !== SocketState.disconnected) return;
		this.idCounter = 0
		this.socket = new SocketAdapter(endpoint)
		this.socket.once('close', this.onSocketClose.bind(this))
		this.socket.on('message', this.onMessage.bind(this))
		await new Promise((res, rej) => {
			this.socket?.once('open', res)
			this.socket?.once('error', rej)
		})
		const resp = await this.unqueuedSend('auth', { token: this.token })
		if(!resp.success) {
			const error = new SatiError('invalid auth token')
			this.setState(SocketState.unrecoverable, error)
			this.socket?.close()
			throw error
		}
		this.setState(SocketState.connected)
		await this.flushQueue()
	}

	public close() {
		this.setState(SocketState.unrecoverable, new SatiError('socket was closed'))
		this.socket?.close()
	}

	private flushQueue() {
		const promises = []
		for(const [ cb, type, data ] of this.queue) {
			promises.push(this.unqueuedSend(type, data).then(cb))
		}
		return Promise.all(promises)
	}

	private onSocketClose() {
		this.socket = null

		for(const awaited of Object.values(this.awaitedMessages)) {
			// we don't want to resend them, because it may cause task double-creation
			awaited.reject(new SatiError('socket gone'))
		}
		this.awaitedMessages = Object.create(null)

		if(this.state !== SocketState.unrecoverable) {
			this.setState(SocketState.disconnected)
			setTimeout(() => {
				this.connect().catch(err => {
					console?.error?.('sati: error while reconnecting:', err)
				})
			}, 1000)
		}
	}

	private async unqueuedSend(type: string, data: any) {
		if(this.state === SocketState.unrecoverable) throw this.error
		if(!this.socket) return;
		const id = ++this.idCounter

		this.socket.send(JSON.stringify({ id, type, data }))

		return new Promise<any>((resolve, reject) => {
			this.awaitedMessages[id] = { resolve, reject }
		})
	}

	public send(type: string, data: any) {
		switch(this.state) {
			case SocketState.connected:
				return this.unqueuedSend(type, data)
			case SocketState.disconnected:
				return new Promise(resolve => {
					this.queue.push([ resolve, type, data ])
				})
			case SocketState.unrecoverable:
				throw this.error
		}
	}

	private	onMessage(message: string) {
		const parsed: WebSocketMessage = JSON.parse(message)
		if(parsed.type === 'event') {
			if(parsed.data.type === 'tokenReissue') {
				this.setState(SocketState.unrecoverable, new SatiError('token was reissued'))
				this.socket?.close()
			}

			this.emit('event', parsed.data)
			return
		}

		if(parsed.to && parsed.to in this.awaitedMessages) {
			this.awaitedMessages[parsed.to].resolve(parsed.data)
		}
	}
}